using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Windows.Controls;
using System.Windows.Threading;
using System.Windows.Media.Imaging;
using System.Windows.Media;
using System.Runtime.InteropServices;
using MediaBridge;

namespace VLC.WPF
{
    public static class VideoOutputController
    {
        private const string name = "WPF_VLC_VIDEO_OUTPUT";
        private const string description = "WPF Video Output Module";

        [DllImport("Kernel32.dll", EntryPoint = "RtlMoveMemory")]
        private static extern void CopyMemory(IntPtr Destination, IntPtr Source, int Length);

        private static Image videoOutput = null;
        private static VideoRendererElement videoRenderer = null;
        private static BitmapBuffer bitmapBuffer = null;
        private static IntPtr bufferPointer = IntPtr.Zero;
        private static Dispatcher dispatcher = Dispatcher.CurrentDispatcher;
        private static bool isInitialized = false;

        private static int videoBufferSize = 0;
        private static bool isInitFrame = false;
        private static int videoOutputType = 0;

        private delegate void InitializeVideoImageEventHandler(int width, int height, float aspect, IntPtr buffer, int bufferSize, int stride);
        private delegate void RenderVideoImageEventHandler(IntPtr buffer);
        private static event InitializeVideoImageEventHandler OnInitializeVideoImage = new InitializeVideoImageEventHandler(InitializeVideo);
        private static event RenderVideoImageEventHandler OnRenderVideoImage = new RenderVideoImageEventHandler(RenderVideoViaImage);

        public delegate void VideoEventHandler();
        public static event VideoEventHandler VideoInitialized;

        private static void OnVideoInitialized()
        {
            if (VideoOutputController.VideoInitialized != null)
            {
                VideoOutputController.VideoInitialized();
            }
        }

        public static string Name
        {
            get { return name; }
        }

        public static string Description
        {
            get { return description; }
        }

        public static Image VideoOutput
        {
            get { return videoOutput; }
            set 
            {
                if (value != null)
                {
                    videoOutput = value;
                    isInitialized = true;
                }
            }
        }

        public static VideoRendererElement VideoRenderer
        {
            get { return videoRenderer; }
            set
            {
                if (value != null)
                {
                    videoRenderer = value;
                    isInitialized = true;
                }
            }
        }

        public static int VideoOutputType
        {
            get { return videoOutputType; }
            set { videoOutputType = value; }
        }

        public static bool IsInitialized
        {
            get { return isInitialized; }
        }

        private static void InitializeVideo(int width, int height, float aspect, IntPtr buffer, int bufferSize, int stride)
        {
            System.Diagnostics.Debug.WriteLine(String.Format("Init: {0}, {1}, {2}, {3}, {4}, {5}", width, height, aspect, buffer, bufferSize, stride));
            if (buffer != IntPtr.Zero && bufferSize > 0)
            {
                if (videoOutput != null)
                {
                    try
                    {
                        BitmapSource bmpsrc = BitmapSource.Create(
                                                width,
                                                height,
                                                96,
                                                96 * aspect * height / width,
                                                PixelFormats.Bgr32,
                                                null,
                                                buffer,
                                                bufferSize,
                                                stride
                                            );
                        videoOutput.Source = bmpsrc;
                        bitmapBuffer = new BitmapBuffer(bmpsrc);
                        bufferPointer = bitmapBuffer.BufferPointer;
                    }
                    catch (Exception e)
                    {
                        System.Diagnostics.Debug.WriteLine("VLC BitmapBuffer Init Error: " + e.Message);
                    }
                }
                if (videoRenderer != null)
                {
                    videoBufferSize = bufferSize;
                    try
                    {
                        videoRenderer.Width = width;
                        videoRenderer.Height = Convert.ToInt32(width / aspect);
                        isInitFrame = true;
                        videoRenderer.TearDown();
                        videoRenderer.InitializeVideo(width, -height, 32);
                        RenderVideoViaRenderer(buffer);
                    }
                    catch (Exception e)
                    {
                        System.Diagnostics.Debug.WriteLine("VLC DirectShow Init Error: " + e.Message);
                    }
                }
                OnVideoInitialized();
            }
        }

        private static void RenderVideoViaRenderer(IntPtr buffer)
        {
            if (buffer != IntPtr.Zero && videoRenderer != null)
            {
                try
                    {
                        videoRenderer.WriteSample(buffer, videoBufferSize, isInitFrame);
                        isInitFrame = true;
                    }
                    catch (Exception e)
                    {
                        System.Diagnostics.Debug.WriteLine("VLC DirectShow Render Error: " + e.Message);
                    }
            }
        }

        private static void RenderVideoViaImage(IntPtr buffer)
        {
            if (buffer != IntPtr.Zero && videoOutput != null && 
                bufferPointer != IntPtr.Zero && bitmapBuffer != null)
            {
                try
                {
                    CopyMemory(bufferPointer, buffer, (int)bitmapBuffer.BufferSize);
                    videoOutput.InvalidateVisual();
                    bufferPointer = bitmapBuffer.BufferPointer;
                }
                catch (Exception e)
                {
                    System.Diagnostics.Debug.WriteLine("VLC BitmapBuffer Render Error: " + e.Message);
                }
            }
        }

        public static void InitializeVideoImage(int width, int height, float aspect, IntPtr buffer, int bufferSize, int stride)
        {
            dispatcher.BeginInvoke(
                    DispatcherPriority.Render,
                    OnInitializeVideoImage,
                    width,
                    height,
                    aspect,
                    buffer,
                    bufferSize,
                    stride
                );
        }

        public static void RenderVideoImage(IntPtr buffer)
        {
            if (videoOutputType == 0)
            {
                dispatcher.BeginInvoke(
                    DispatcherPriority.Render,
                    OnRenderVideoImage,
                    buffer
                );
            }
            else if (videoOutputType == 1)
            {
                RenderVideoViaRenderer(buffer);
            }
        }
    }
}
