﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Threading;
using System.ComponentModel;

namespace VLC.WPF
{
    internal enum VlcWatcherState
    {
        Started,
        Running,
        Paused,
        Holding,
        Stopped
    }

    public class VlcWatcher : INotifyPropertyChanged
    {
        private const float VIDEO_END_POSITION = -20.0f;

        private VlcWatcherState state = VlcWatcherState.Stopped;

        private Dispatcher watcher = null;
        private DispatcherTimer watcherTimer = null;
        private VlcElement vlcElement = null;

        private string videoTimeStr;
        private bool videoIsPlaying;
        private float videoPosition;
        private int videoLength;
        private TimeSpan videoPositionTime;
        private TimeSpan videoLengthTime;

        public VlcWatcher(VlcElement vlcElement)
        {
            this.vlcElement = vlcElement;
            watcher = Dispatcher.CurrentDispatcher;
            watcherTimer = new DispatcherTimer(new TimeSpan(0, 0, 0, 0, 500), DispatcherPriority.Normal, new EventHandler(WatcherRefresh), watcher);
            watcherTimer.Stop();
            Reset();
        }

        private void OnWatcherUpdated()
        {
            if (WatcherUpdated != null)
            {
                WatcherUpdated(this, EventArgs.Empty);
            }
        }

        private void OnContentFinished()
        {
            if (ContentFinished != null)
            {
                ContentFinished(this, EventArgs.Empty);
            } 
        }

        internal void OnContentStarted()
        {
            if (ContentStarted != null)
            {
                ContentStarted(this, EventArgs.Empty);
            }
        }

        private void WatcherRefresh(object sender, EventArgs e)
        {
            VideoIsPlaying = vlcElement.IsPlaying;
            if (state == VlcWatcherState.Started && VideoIsPlaying) state = VlcWatcherState.Running;
            if (state == VlcWatcherState.Paused && VideoIsPlaying) state = VlcWatcherState.Running;
            if (state == VlcWatcherState.Running || state == VlcWatcherState.Paused)
            {
                VideoPosition = vlcElement.Position;
                VideoLength = vlcElement.Length;
                if (VideoPosition == VIDEO_END_POSITION)
                {
                    OnContentFinished();
                    Stop();
                }
                else
                {
                    try
                    {
                        VideoPositionTime = VlcWatcher.GenerateTimeSpan(Convert.ToInt32(VideoPosition * VideoLength));
                    }
                    catch
                    {
                        VideoPositionTime = TimeSpan.Zero;
                    }
                    VideoLengthTime = VlcWatcher.GenerateTimeSpan(VideoLength);
                    VideoTimeStr = VideoPositionTime.ToString() + " / " + VideoLengthTime.ToString();
                    OnWatcherUpdated();
                }                
            }
        }

        private void Reset()
        {
            VideoTimeStr = "--:--:-- / --:--:--";
            VideoIsPlaying = false;
            VideoPosition = 0.0f;
            VideoLength = 0;
            VideoPositionTime = TimeSpan.Zero;
            VideoLengthTime = TimeSpan.Zero;
        }

        internal void Start()
        {
            if (state != VlcWatcherState.Paused)
            {
                Reset();
                state = VlcWatcherState.Started;
                watcherTimer.Start();
                OnWatcherUpdated();
            }
        }

        internal void Pause()
        {
            state = VlcWatcherState.Paused;
        }

        internal void Stop()
        {
            Reset();
            state = VlcWatcherState.Stopped;
            watcherTimer.Stop();
            OnWatcherUpdated();
        }

        public event EventHandler WatcherUpdated;
        public event EventHandler ContentStarted;
        public event EventHandler ContentFinished;        

        public void HoldOn()
        {
            if (state == VlcWatcherState.Running)
            {
                state = VlcWatcherState.Holding;
            }
        }

        public void Resume()
        {
            if (state == VlcWatcherState.Holding)
            {
                state = VlcWatcherState.Running;
            }
        }

        public static TimeSpan GenerateTimeSpan(int secs)
        {
            if (secs > 0)
            {
                return new TimeSpan(
                    secs / 3600,
                    (secs % 3600) / 60,
                    (secs % 3600) % 60
                    );
            }
            else
            {
                return new TimeSpan(0, 0, 0);
            }
        }

        public TimeSpan GenerateTimeSpan(float videoPosition)
        {
            return VlcWatcher.GenerateTimeSpan(VideoPositionToSecs(videoPosition));
        }

        public int VideoPositionToSecs(float videoPosition)
        {
            return Convert.ToInt32(videoPosition * VideoLength);
        }

        public string VideoTimeStr
        {
            get { return videoTimeStr; }
            internal set
            {
                videoTimeStr = value;
                OnPropertyChanged("VideoTimeStr");
            }
        }

        public bool VideoIsPlaying
        {
            get { return videoIsPlaying; }
            internal set
            {
                videoIsPlaying = value;
                OnPropertyChanged("VideoIsPlaying");
            }
        }

        public float VideoPosition
        {
            get { return videoPosition; }
            internal set
            {
                videoPosition = value;
                OnPropertyChanged("VideoPosition");
            }
        }

        public int VideoLength
        {
            get { return videoLength; }
            internal set
            {
                videoLength = value;
                OnPropertyChanged("VideoLength");
            }
        }

        public TimeSpan VideoPositionTime
        {
            get { return videoPositionTime; }
            internal set
            {
                videoPositionTime = value;
                OnPropertyChanged("VideoPositionTime");
            }
        }

        public TimeSpan VideoLengthTime
        {
            get { return videoLengthTime; }
            internal set
            {
                videoLengthTime = value;
                OnPropertyChanged("VideoLengthTime");
            }
        }

        #region INotifyPropertyChanged Member

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #endregion
    }
}
